#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>
#include "poly.h"
#include "basic.h"
#include "pool.h"
#include "monitor.h"


static int verbose = 0 ;

uintmax_t add(uintmax_t y,uintmax_t r) { return y+r; }

/* Common */

typedef struct {
  pool_t *pool ;
  monitor_t *monitor ;
} common_t ;

/* Worker */
typedef struct {
  subtask_t *arg ;
  common_t *common ;
} worker_t ;


worker_t *alloc_worker_t(subtask_t *arg,common_t *c) {
  worker_t *p = malloc_check(sizeof(*p)) ;
  p->arg = copy_arg(arg) ;
  p->common = c ;
  return p ;
}

void free_worker_t(worker_t *p) {
  free_arg(p->arg) ;
  free(p);
}

void *worker(void *p) {
  worker_t *w = (worker_t *)p ;
  common_t *c = w->common ;
  subtask_t *arg = w->arg ;
  count_t y = run_subtask(arg) ;
  free_worker_t(w) ;
  leave_pool(c->pool) ;
  leave_monitor(c->monitor,y) ;
  return NULL ;
}


void handle_subtask(void *env,subtask_t *z) {
  if (verbose > 1) {
    fprintf(stderr,"SUBTASK: ") ;
    print_subtask(stderr,z) ;
  }
  common_t *c = env ;
  look_pool(c->pool) ;
  enter_monitor(c->monitor) ;
  worker_t *w = alloc_worker_t(z,c) ;
  create_thread_detached(worker,w) ;
}

count_t run(int n, int depth, int nprocs) {
  /* Initialise */
  common_t c ;
  c.monitor = alloc_monitor(add,0) ;
  c.pool = alloc_pool(nprocs) ;
  /* generate subtasks */
  int ntasks =  generate_subtasks(n, depth, &c, handle_subtask) ;
  if (verbose) {
    fprintf(stderr,"Generated %i subtasks\n",ntasks) ;
  }
  /* wait result */
  count_t r =  wait_monitor(c.monitor,ntasks) ;
  /* clean */
  free_pool(c.pool) ;
  free_monitor(c.monitor) ;
  return r ;
}

void usage(char *pgm)
{
  fprintf(stderr,"usage: %s [-v] [-d<int>] [-j<int>]*\n",pgm);
  exit(-1);
}
int main(int argc, char **argv) {
  char *arg=0;
  char *pgm = argv[0];
  int n = 15 ;
  int depth = 5 ;
  int nprocs = 4 ;

  while (--argc) {
    arg = *++argv;
    if (*arg == '-') {
      arg++ ;
       if (*arg == 'd') {
        if (arg[1]) {
          depth = atoi(arg+1) ;
        } else {
          usage(pgm) ;
        }
       } else if (*arg == 'p' || *arg == 'j') {
         if (arg[1]) {
           nprocs = atoi(arg+1) ;
         } else {
           usage(pgm) ;
         }
       } else if (*arg == 'v') {
         verbose++;
       } else {
         usage(pgm);
       }
    } else break ;      
  }
  if (argc >= 1) n = atoi(arg);
  if (n > NMAX) {
      fprintf(stderr,"argument must be <= %d\n",NMAX);
      exit(2);
  }

  count_t r = run(n,depth,nprocs) ;
  printf("%" PRC "\n",r) ;
  return 0 ;
}
