#include <stdlib.h>
#include <stdio.h>
#include "basic.h"
#include "pool.h"
#include "monitor.h"


static int verbose = 0 ;

int compute(int x) {
  if (verbose) fprintf(stderr,"Compute: %i\n",x) ;
  return x ;
}

uintmax_t add(uintmax_t y,uintmax_t r) { return y+r; }


/* Common */

typedef struct {
  pool_t *pool ;
  monitor_t *monitor ;
} common_t ;

/* Worker */
typedef struct {
  int arg ;
  common_t *common ;
} worker_t ;

worker_t *alloc_worker_t(int arg,common_t *c) {
  worker_t *p = malloc_check(sizeof(*p)) ;
  p->arg = arg ;
  p->common = c ;
  return p ;
}

void free_worker_t(worker_t *p) { free(p); }

void *worker(void *p) {
  worker_t *w = (worker_t *)p ;
  common_t *c = w->common ;
  int arg = w->arg ;
  free_worker_t(w) ;
  int y = compute(arg) ;
  leave_pool(c->pool) ;
  leave_monitor(c->monitor,y) ;
  return NULL ;
}


void spawn_worker(int arg, common_t *c) {
  worker_t *w = alloc_worker_t(arg,c) ;
  create_thread_detached(worker,w) ;
}

void master(int nprocs, int n) {
  common_t c ;
  c.monitor = alloc_monitor(add,0) ;
  c.pool = alloc_pool(nprocs) ;
  for (int k = 1 ; k <= n ; k++) {
    look_pool(c.pool) ;
    enter_monitor(c.monitor) ;
    spawn_worker(k,&c) ;
  }
  int r = wait_monitor(c.monitor,n) ;
  free_pool(c.pool) ;
  free_monitor(c.monitor) ;
  printf("%i\n",r) ;
}


void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [nprocs] [n]\n",pgm) ;
}

int main(int argc, char **argv) {
  char * pgm = argv[0] ;
  int nprocs = 4 ;
  int n = 100 ;
  ++argv ;
  --argc ;
  while (argc > 0) {
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    ++argv;
    --argc ;
  }
  if (argc == 1) {    
    nprocs = atoi(argv[0]) ;
  } else if (argc == 2) {
    nprocs = atoi(argv[0]) ;
    n = atoi(argv[1]) ;
  }
  master(nprocs,n) ;
  return 0 ;
}
