/********************************************************************/
/*                              Poly                                */
/*                                                                  */
/*   Luc Maranget, projet Moscova, Inria Rocquencourt               */
/*                                                                  */
/* Copyright 2005 Institut National de Recherche en Informatique et */
/* en Automatique.  All rights reserved.  This file is distributed  */
/* under the terms of the Q Public License version 1.0.             */ 
/********************************************************************/

#include "poly.h"

typedef enum statut {libre, plein} statut;

typedef struct point {
  statut stat;
  struct point *next;
  struct point *voisins[5];
} point;

/* damier */

typedef struct {
  int n ;
  point t[NMAX][2*(NMAX-1)];
} board_t ;


int check(board_t *b,int i,int j)
{
  int n = b->n ;

  if ((0 <= i) && (i < n) && 0 <= j && (j < 2*(n-1))) {
    return ((i > 0) || (j > n-2));
  } else {
   return 0;
  }
}

void set_nexts(board_t *b,int i,int j) {
  int k=0;
  point *r = &b->t[i][j];

  if (check(b,i+1,j)) {r->voisins[k++] = &b->t[i+1][j];}
  if (check(b,i,j+1)) {r->voisins[k++] = &b->t[i][j+1];}
  if (check(b,i-1,j)) {r->voisins[k++] = &b->t[i-1][j];}
  if (check(b,i,j-1)) {r->voisins[k++] = &b->t[i][j-1];}
  r->voisins[k] = 0;
}

void init_damier(board_t *b)
{
  int i,j;
  int n = b->n ;
  for (i=0;i<n;i++) {
    for (j=0;j<2*(n-1);j++) {
      set_nexts(b,i,j);
    }
  }
}

board_t *alloc_board(int n) {
  board_t *r = calloc(sizeof(*r),1) ;
  r->n = n ;
  init_damier(r) ;
  return r ;
}

void free_board(board_t *p) { free(p) ; }

void print_sol(FILE *fp,board_t *b)
{
  int i,j;
  point *r;
  int n = b->n ;

  for (i=n-1;i >= 0;i--) {
    for (j=0;j<2*(n-1);j++) {
      r = &b->t[i][j];
      fputc(r->stat==libre?'.':(r->stat==plein?'#':'X'),fp);
    }
    putc('\n',fp);
  }
  putc('\n',fp);
}

/* Fast enumerator */
count_t do_search(int k,point *nexts,unsigned int l, board_t *b) {
  point *news,**v,*me;
  count_t r = 0 ;
  if (k > 1) {
    
    for (;nexts;) {
      v = &(nexts->voisins[0]);
      nexts = nexts->next; l--;
      news = nexts;
      for (me = *v ; me ; me = *++v) {
        if ((me)->stat == libre) {
          (me)->stat = plein;
          (me)->next = news;
          news = me; l++;
        }
      }
      r += do_search(k-1,news,l,b);
      for (; news != nexts; news = news->next) {
        news->stat = libre; l--;
      }
    }
  } else {
    l = l * (l-1)/2;
    for (;nexts;) {
      v = &(nexts->voisins[0]);
      nexts = nexts->next;
      for (me = *v ; me ; me = *++v) {
        if (me->stat == libre)
          l++;
      }
    }
    r += l;
  }
  return r ;
}


void error_read(char *msg) {
  fprintf(stderr, "Bad checkpoint file %s: %s\n", "(stdin)", msg) ;
  exit(2) ;
}

void print_subtask(FILE *fp, subtask_t *z) {
  fprintf(fp,"%i", z->id) ;
  fprintf(fp," %i", z->n) ;
  fprintf(fp," %i", z->depth) ;
  for (int k = z->n-1 ; k >= z->depth ; k--) {
    fprintf(fp, " %i", z->seen[k]) ;
  }
  fprintf(fp, "\n") ;
  fflush(fp) ;
}

void free_arg(subtask_t *z) { return free(z) ; }

subtask_t *alloc_arg(int id, int n, int depth) {
  subtask_t *r = calloc(sizeof(*r),1) ;
  if (!r) {
    perror("calloc") ;
    exit(2) ;
  }
  r->id = id ;
  r->n = n ;
  r->depth = depth ;
  return r ;
}

subtask_t *copy_arg(subtask_t *z) {
  subtask_t *r = calloc(sizeof(*r),1) ;
  if (!r) {
    perror("calloc") ;
    exit(2) ;
  }
  r->id = z->id ;
  r->n = z->n ;
  r->depth = z->depth ;
  for (int k = 0 ; k < NMAX ; k++) r->seen[k] = z->seen[k] ;
  return r ;
}

subtask_t *read_subtask(FILE *fp) {  
  int id,n,depth ;  
  if (fscanf(fp, "%i", &id) != 1) error_read("id") ;
  if (fscanf(fp, "%i", &n) != 1) error_read("n") ;
  if (fscanf(fp, "%i", &depth) != 1) error_read("depth") ;
  subtask_t *r = alloc_arg(id,n,depth) ;
  /* Seen array */
  for (int k = n-1 ; k >= r->depth ; k--) {
    if (fscanf(fp, "%i", &r->seen[k]) != 1) error_read("seen[k]") ;
  }
  return r ;
}





int subgen(int k, point *nexts, subtask_t *z, board_t *b, void *env, emit_t emit) {
  int r = 0 ;
  if (k <= z->depth) {
    for ( ; nexts ; nexts = nexts->next) {
      r++ ;
      z->seen[k]++ ;
      z->id++ ;
      z->depth = k ;
      emit(env,z) ;
    }
  } else {
    for ( ; nexts ; ) {
      point *me, *news ;
      point **v = &(nexts->voisins[0]) ;
      z->seen[k]++ ;
      nexts = nexts->next ;
      news = nexts ;
      for (me = *v ; me ; me = *++v) {
        if (me->stat == libre) {
          me->stat = plein ;
          me->next = news ;
          news = me ;
        }
      }
      r += subgen(k-1, news, z, b, env, emit) ;
      for ( ; news != nexts ; news = news->next) {
        news->stat = libre ;
      }
    }
  }
  z->seen[k] = 0 ;
  return r ;
}

int generate_subtasks(int n, int depth, void *env,emit_t emit) {
  int d = n-depth ;
  if (d < 2) d = 2 ;
  board_t *b = alloc_board(n) ;
  point *p=&b->t[0][n > 1 ? n-2 : 0];
  p->next = 0; p->stat = plein;
  subtask_t *z = alloc_arg(0,n,d) ;
  return subgen(n-1,p,z,b,env,emit) ;
}

/* Execute subtask */
count_t substart(int k, point *nexts, int size, subtask_t *z, board_t *b) {
  int s ;
  point *me, **v ;

  for (s = z->seen[k] ; s > 1 ; s--) {
    nexts = nexts->next ; size-- ;
  }
  me = nexts ; nexts = nexts->next ; size-- ;
  for (v = &(me->voisins[0]) ; *v ; v++) {
    if ((*v)->stat == libre) {
      (*v)->stat = plein ;
      (*v)->next = nexts ;
      nexts = *v ; size++ ;
    }
  }
  if (k > z->depth) {
    return substart(k-1, nexts, size, z, b) ;
  } else {
    return do_search(k-1, nexts, size, b) ;
  }
}


/* Execute a subtask, with board */
static count_t run_subtask_board(subtask_t *z, board_t *b) {
  int n = z->n ;
  point *p=&b->t[0][n > 1 ? n-2 : 0];
  p->next = 0; p->stat = plein;
  return substart(n-1,p,1,z,b);
}

/* Execute a subtask */
count_t run_subtask(subtask_t *z) {
  board_t *b = alloc_board(z->n) ;
  count_t r =  run_subtask_board(z,b) ;
  free_board(b) ;
  return r ;
}

/* Execute a complete computation */
count_t run_task(int n) {
  board_t *b = alloc_board(n) ;
  point *p=&b->t[0][n > 1 ? n-2 : 0];
  p->next = 0; p->stat = plein;
  return do_search(n-1,p,1,b) ;
}
